/**
 ** MTIMTITaskCodels.cc
 **
 ** Codels called by execution task MTIMTITask
 **
 ** Author: 
 ** Date: 
 **
 **/

#ifdef VXWORKS
# include <vxWorks.h>
#else
# include <portLib.h>
#endif
#include "server/MTIHeader.h"

#include "h2mathLib.h"
#include "h2evnLib.h"
#include "h2timeLib.h"

#include <MTI-clients/structMTI.h>
#include <MTI-clients/MTI.h>
#include <math.h>

#include <stdio.h>

MTI * MTIClient = NULL;

static int nbSamplesUsed;
static double meanRoll, meanPitch;
static double varianceRoll, variancePitch;

static timeval tv1, tv2;
static timezone tz1, tz2;

/*------------------------------------------------------------------------
 *
 * MTIInitExec : configure the MTI device and connect to it
 *
 * Description: 
 * 
 * Returns:    ETHER
 */

ACTIVITY_EVENT
MTIInitExec(INERTIAL_DEVICE *device, int *report)
{
	/* Serial speed and COM to use */
	//strcpy(SDI_F->config.device,device);	 /* serial port number : COM1, COM2 ...*/

	// DATA OUTPUT MODES AND FORMATS
	SDI_F->config.outputMode   = MTI_OPMODE_BOTH;
	SDI_F->config.outputFormat = MTI_OPFORMAT_EULER;

	// Start MTI with default settings
	MTIClient  = new MTI(SDI_F->device.device,
			     SDI_F->config.outputMode,
			     SDI_F->config.outputFormat);


	// DEFAULT SYNCOUT PIN MODES AND SETTINGS
	SDI_F->config.syncOutMode          = MTI_SYNCOUTMODE_DISABLED;
	SDI_F->config.syncOutPulsePolarity = MTI_SYNCOUTPULSE_POS;

	// number of acquisitions to skip before syncOut pin actuates
	SDI_F->config.syncOutSkipFactor = 0; // mark all acquisitions

	// number of clock ticks @ 33.9ns to offset pin action from sensor sampling
	SDI_F->config.syncOutOffset = 0; // no offset

	// number of clock ticks @ 33.9ns to define pulse width
	SDI_F->config.syncOutPulseWidth = 29498; // 1 ms pulse

	// Set SyncOut settings
	MTIClient->set_syncOut(SDI_F->config.syncOutMode,
			       SDI_F->config.syncOutPulsePolarity,
			       SDI_F->config.syncOutSkipFactor,
			       SDI_F->config.syncOutOffset,
			       SDI_F->config.syncOutPulseWidth);
			       
	/* Sets the biases to 0.0 */
	SDI_F->biases.pitchBias = 0.0;
	SDI_F->biases.rollBias  = 0.0;

	return ETHER;
}

/*------------------------------------------------------------------------
 *
 * MTIConfigExec : configure the MTI device and connect to it
 *
 * Description: 
 * 
 * Returns:    ETHER
 */

ACTIVITY_EVENT
MTIConfigExec(INERTIAL_CONFIG *config, int *report)
{

	MTIClient->set_mode(SDI_F->config.outputMode);
		
	MTIClient->set_outputDisplay(SDI_F->config.outputFormat);
	
	MTIClient->set_syncOut(SDI_F->config.syncOutMode,
			       SDI_F->config.syncOutPulsePolarity,
			       SDI_F->config.syncOutSkipFactor, 
			       SDI_F->config.syncOutOffset, 
			       SDI_F->config.syncOutPulseWidth);
	
	/* Sets the biases to 0.0 */
	SDI_F->biases.pitchBias = 0.0;
	SDI_F->biases.rollBias = 0.0;

	return ETHER;
}

/*------------------------------------------------------------------------
 *
 * MTIResetExec : Reset the MTI
 *
 * Description: 
 * 
 * Returns:    ETHER
 */

ACTIVITY_EVENT
MTIResetExec(int *report)
{

	MTIClient->reset();
		
	return ETHER;
}

/*------------------------------------------------------------------------------
 * MTIExecTaskPerm : codels of permanent activity
 *
 * Read the data exported by MTI, and export them to pom
 *
 * Returns : OK
 */ 
STATUS
MTIExecTaskPerm(int * report)
{
	if ( MTIClient == NULL || !MTIClient->is_connected() )
		return OK;

	// read sensor	
	gettimeofday (&tv1, &tz1);
	MTIClient->read(&(SDI_F->sensorData), false);
	//	gettimeofday (&tv, &tz);

	// Write it into the Pom structure
	POM_SENSOR_POS* pom = &(SDI_F->pomPos);

	POM_EULER current;
	memset(&current, 0, sizeof(current));

	/*
	 * XXX The yaw data is mostly not usable in the current environnement due to
	 * magnetic disturbance. We need to know how to calibrate it to have a
	 * correct yaw.
	 * Pitch and roll are correct, so it is ok to use for 3D odometry (with
	 * rflex)
	 */
	current.yaw   = 
	  M_PI * SDI_F->sensorData.EULER[2] / 180.0;
	current.pitch = 
	  M_PI * (SDI_F->sensorData.EULER[1] - SDI_F->biases.pitchBias) / 180.0;
	current.roll  = 
	  M_PI * (SDI_F->sensorData.EULER[0] - SDI_F->biases.rollBias) / 180.0;
	pom->sensorToMain.euler = current;
	return OK;
}

/*------------------------------------------------------------------------------
 * MTIExecTaskEnd : Exit handler
 *
 * When the Exec Task finish, just kill the MTIClient
 * MTIClient dtor do all the necessary stuff to stop gracefully the process
 *
 * Returns : OK
 */
STATUS
MTIExecTaskEnd()
{
	delete MTIClient;
	return OK;
}


/*------------------------------------------------------------------------
 * Log
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MTI_CANNOT_OPEN_FILE
 *              S_MTI_CANNOT_WRITE_FILE
 */

static FILE *MTILogFile;

/* MTILogStart  -  codel START of Log
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MTILogStart(GEN_STRING128 *fileName, int *report)
{
  if (!(MTILogFile = fopen(fileName->name, "w"))) {
    printf ("MTI: cannot open odometer file %s\n", fileName->name);
    *report = S_MTI_CANNOT_OPEN_FILE;
    return ETHER;
  }

  printf ("MTILogStart: log file %s found\n", fileName->name);
  fprintf (MTILogFile, "# Time (sec) Count "
	   "eulerX eulerY eulerZ (deg) "
	   "gyroX gyroY gyroZ (rad/s) "
	   "accX accY accZ (m/s2) "
	   "magX magY magZ\n");
  return EXEC;
}

/* MTILogExec  -  codel EXEC of Log
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MTILogExec(GEN_STRING128 *fileName, int *report)
{
  /*  H2TIMESPEC tv; */
  INERTIAL_DATA data;

  /* get date */
  /*  h2GetTimeSpec(&tv);*/

  /* save date */
  if (!fprintf(MTILogFile, "%lf ", 
	       tv1.tv_sec + tv1.tv_usec*1.0E-6)) {
    printf ("MTI: cannot log in %s\n", fileName->name);
    MTILogFile = 0;
    *report = S_MTI_CANNOT_WRITE_FILE;
    return ETHER;
  }

  /* save POM_ME_POS */
  /*  POM_EULER current = SDI_F->pomPos.sensorToMain.euler; */
  /* fprintf(MTILogFile, "yaw = %g\n", current.yaw); */
  /* fprintf(MTILogFile, "pitch = %g\n", current.pitch); */
  /* fprintf(MTILogFile, "roll = %g\n", current.roll); */

  data = SDI_F->sensorData;
  fprintf (MTILogFile, "%d %f %f %f %f %f %f %f %f %f %f %f %f\n",
	   data.COUNT,
	   data.EULER[0], data.EULER[1], data.EULER[2],
	   data.GYR[0], data.GYR[1], data.GYR[2],
	   data.ACC[0], data.ACC[1], data.ACC[2],
	   data.MAG[0], data.MAG[1], data.MAG[2]);

//   fflush(MTILogFile); 

  return EXEC;
}

/* MTILogInter  -  codel END of Log
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MTILogInter(GEN_STRING128 *fileName, int *report)
{
  fclose(MTILogFile);
  MTILogFile = NULL;

  return ETHER;
}

/*------------------------------------------------------------------------
 * AverageSamples
 *
 * Description: Computes the mean and standard deviation of the roll and pitch
 * angles  
 *
 * Reports:     OK
 *              S_MTI_NOT_CONNECTED
 *              S_MTI_NEGATIVE_NB_SAMPLES
 *
 */

/* MTIAverageStart - codel Start of AverageSamples
   Returns:  ETHER EXEC */
ACTIVITY_EVENT
MTIAverageStart (AVERAGING_PARAMETERS* parameter, int *report)
{
  if ( MTIClient == NULL || !MTIClient->is_connected() ) {
    printf ("MTI: not connected\n");
    *report = S_MTI_NOT_CONNECTED;
    return ETHER;
  }
  if (parameter->nbSamples <= 0) {
    printf ("MTI: enter a number of samples > 0 !\n");
    *report = S_MTI_NEGATIVE_NB_SAMPLES;
    return ETHER;
  }

  nbSamplesUsed = 0;

  return EXEC;
}

/* MTIAverageDo - main codel of AverageSamples
   Returns:  ETHER EXEC */
ACTIVITY_EVENT
MTIAverageDo (AVERAGING_PARAMETERS* parameter, int *report)
{
  double previousMeanPitch, previousMeanRoll;

  if (nbSamplesUsed == 0) {
    /* Initialize the means and variances */
    meanPitch = SDI_F->sensorData.EULER[1];
    meanRoll = SDI_F->sensorData.EULER[0];
    variancePitch = 0.0;
    varianceRoll = 0.0;
    nbSamplesUsed++;
  }
  else {
    if (nbSamplesUsed < parameter->nbSamples) {
      previousMeanPitch = meanPitch;
      meanPitch = ((previousMeanPitch * nbSamplesUsed) 
		   + SDI_F->sensorData.EULER[1])
	/ (nbSamplesUsed + 1);
      variancePitch += (SDI_F->sensorData.EULER[1] - previousMeanPitch) 
	* (SDI_F->sensorData.EULER[1] - meanPitch);
      
      previousMeanRoll = meanRoll;
      meanRoll = ((previousMeanRoll * nbSamplesUsed) 
		  + SDI_F->sensorData.EULER[0])
	/ (nbSamplesUsed + 1);
      varianceRoll += (SDI_F->sensorData.EULER[0] - previousMeanRoll) 
	* (SDI_F->sensorData.EULER[0] - meanRoll);
      nbSamplesUsed++;
    }
    else { // nbSamplesUsed = nbSamples, we are done
      return END;
    }
  }
  return EXEC;
}

/* MTIAverageEnd - main codel of AverageSamples
   Returns:  ETHER */
ACTIVITY_EVENT
MTIAverageEnd (AVERAGING_PARAMETERS* parameter, int *report)
{
  printf ("Pitch: mean = %.4f, std deviation = %.8f (on %d samples)\n",
	  meanPitch, sqrt (variancePitch/nbSamplesUsed), nbSamplesUsed);
  printf ("Roll: mean = %.4f, std deviation = %.8f (on %d samples)\n",
	  meanRoll, sqrt (varianceRoll/nbSamplesUsed), nbSamplesUsed);
  return ETHER;
}

/*------------------------------------------------------------------------
 * SetBiases
 *
 * Description: Sets the pitch and roll angle biases to the specified values
 * angles  
 *
 * Reports:     OK
 *              S_MTI_NOT_CONNECTED
 *
 */

/* MTIAverageStart - codel Start of AverageSamples
   Returns:  ETHER EXEC */
ACTIVITY_EVENT
MTISetBiases (MTI_BIASES* biases, int *report)
{
  if ( MTIClient == NULL || !MTIClient->is_connected() ) {
    printf ("MTI: not connected\n");
    *report = S_MTI_NOT_CONNECTED;
    return ETHER;
  }
  SDI_F->biases.pitchBias = biases->pitchBias;
  SDI_F->biases.rollBias = biases->rollBias;

  return ETHER;
}
